# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 TU Wien.
#
# Invenio-Keycloak is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pre-defined defaults and helpers for Invenio-Keycloak configuration."""

from invenio_oauthclient.contrib.settings import OAuthSettingsHelper


class KeycloakSettingsHelper(OAuthSettingsHelper):
    """Helper for creating REMOTE_APP configuration dictionaries for Keycloak.

    This class can be used to easily create a base configuration with sensible
    defaults for a default-ish Keycloak setup.
    It requires knowledge about the base URL of the Keycloak instance and the
    realm on which the Invenio client application is configured.
    Because the default endpoint URLs follow a simple schema, this information
    can be used to create a simple base configuration.

    The methods ``remote_app()`` and ``remote_rest_app()`` create and return
    a dictionary in the form expected by Invenio-OAuthClient.
    The latter can be used for providing SSO capabilities to SPAs communicating
    with Invenio via the REST API.

    Further, the helper provides some properties for commonly used default
    endpoint URLs.
    """

    def __init__(
        self,
        title,
        description,
        base_url,
        realm,
        app_key=None,
        icon=None,
        scopes="openid",
        **kwargs
    ):
        """The constructor takes two arguments.

        :param base_url: The base URL on which Keycloak is running
                            (e.g. "http://localhost:8080")
        :param realm: Realm in which the invenio client application is defined
        """
        app_key = app_key or "KEYCLOAK_APP_CREDENTIALS"
        base_url = "{}/".format(base_url.rstrip("/"))  # add leading `/`

        self._realm_url = "{}auth/realms/{}".format(base_url, realm)

        access_token_url = self.make_url(self._realm_url, "token")
        authorize_url = self.make_url(self._realm_url, "auth")
        self._user_info_url = self.make_url(self._realm_url, "userinfo")

        super().__init__(
            title,
            description,
            base_url,
            app_key,
            icon=icon,
            request_token_params={"scope": scopes},
            access_token_url=access_token_url,
            authorize_url=authorize_url,
            **kwargs
        )

        self._handlers = dict(
            authorized_handler="invenio_oauthclient.handlers:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.keycloak.handlers:disconnect_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.keycloak.handlers:info_handler",
                info_serializer="invenio_oauthclient.contrib.keycloak.handlers:info_serializer_handler",
                setup="invenio_oauthclient.contrib.keycloak.handlers:setup_handler",
                view="invenio_oauthclient.handlers:signup_handler",
            ),
        )

        self._rest_handlers = dict(
            authorized_handler="invenio_oauthclient.handlers.rest:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.keycloak.handlers:disconnect_rest_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.keycloak.handlers:info_handler",
                info_serializer="invenio_oauthclient.contrib.keycloak.handlers:info_serializer_handler",
                setup="invenio_oauthclient.contrib.keycloak.handlers:setup_handler",
                view="invenio_oauthclient.handlers.rest:signup_handler",
            ),
            response_handler=(
                "invenio_oauthclient.handlers.rest:default_remote_response_handler"
            ),
            authorized_redirect_url="/",
            disconnect_redirect_url="/",
            signup_redirect_url="/",
            error_redirect_url="/",
        )

    @property
    def user_info_url(self):
        """URL for the user info endpoint."""
        return self._user_info_url

    @property
    def realm_url(self):
        """URL for the realm's endpoint."""
        return self._realm_url

    @staticmethod
    def make_url(realm_url, endpoint):
        """Create an endpoint URL following the default Keycloak URL schema.

        :param realm_url: The realm base URL
        :param endpoint: The endpoint to use (e.g. "auth", "token", ...)
        """
        return "{}/protocol/openid-connect/{}".format(realm_url, endpoint)

    def get_handlers(self):
        """Return a dict with the auth handlers."""
        return self._handlers

    def get_rest_handlers(self):
        """Return a dict with the auth REST handlers."""
        return self._rest_handlers



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Views for OAuth."""
from flask import url_for
from flask_admin import Admin
from invenio_db import db

from invenio_oauthclient import InvenioOAuthClient
from invenio_oauthclient.admin import (
    remote_account_adminview,
    remote_token_adminview,
    user_identity_adminview,
)


def test_admin(app):
    """Test flask-admin interace."""
    InvenioOAuthClient(app)

    assert isinstance(remote_account_adminview, dict)
    assert isinstance(remote_token_adminview, dict)
    assert isinstance(user_identity_adminview, dict)

    assert "model" in remote_account_adminview
    assert "modelview" in remote_account_adminview
    assert "model" in remote_token_adminview
    assert "modelview" in remote_token_adminview
    assert "model" in user_identity_adminview
    assert "modelview" in user_identity_adminview

    admin = Admin(app, name="Test")

    user_model = remote_account_adminview.pop("model")
    user_view = remote_account_adminview.pop("modelview")
    admin.add_view(user_view(user_model, db.session, **remote_account_adminview))

    with app.app_context():
        # create user and save url for testing
        request_url = url_for("remoteaccount.index_view")

    with app.app_context():
        with app.test_client() as client:
            res = client.get(request_url, follow_redirects=True)
            assert res.status_code == 200
            assert "Extra Data" in str(res.get_data())
            assert "Tokens" in str(res.get_data())



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test handlers."""

import pytest

from invenio_oauthclient.errors import OAuthResponseError
from invenio_oauthclient.handlers import response_token_setter, token_getter
from invenio_oauthclient.models import RemoteToken
from invenio_oauthclient.utils import oauth_authenticate


def test_token_setter(app, remote):
    """Test token setter on response from OAuth server."""

    # OAuth1
    resp_oauth1 = {
        "name": "Josiah Carberry",
        "expires_in": 3599,
        "oauth_token": "test_access_token",
        "oauth_token_secret": "test_refresh_token",
        "scope": "/authenticate",
        "token_type": "bearer",
    }
    assert not response_token_setter(remote, resp_oauth1)

    # Bad request
    resp_bad = {
        "invalid": "invalid",
    }
    with pytest.raises(OAuthResponseError):
        response_token_setter(remote, resp_bad)


def test_token_getter(remote, models_fixture, app):
    """Test token getter on response from OAuth server."""
    datastore = app.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    # Missing RemoteToken
    oauth_authenticate("dev", user)
    assert not token_getter(remote)

    # Populated RemoteToken
    RemoteToken.create(user.id, "testkey", "mytoken", "mysecret")
    oauth_authenticate("dev", user)
    assert token_getter(remote) == ("mytoken", "mysecret")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from .rest import authorized_default_handler as authorized_default_handler_rest
from .rest import authorized_signup_handler as authorized_signup_handler_rest
from .rest import disconnect_handler as disconnect_handler_rest
from .rest import signup_handler as signup_handler_rest
from .ui import (
    authorized_default_handler,
    authorized_signup_handler,
    disconnect_handler,
    oauth2_handle_error,
)
from .ui import oauth_resp_remote_error_handler as oauth_error_handler
from .ui import signup_handler
from .utils import (
    authorized_handler,
    get_session_next_url,
    make_handler,
    make_token_getter,
    oauth1_token_setter,
    oauth2_token_setter,
    oauth_logout_handler,
    response_token_setter,
    set_session_next_url,
    token_delete,
    token_getter,
    token_session_key,
    token_setter,
)

__all__ = (
    "authorized_default_handler_rest",
    "authorized_default_handler",
    "authorized_handler",
    "authorized_signup_handler_rest",
    "authorized_signup_handler",
    "disconnect_handler_rest",
    "disconnect_handler",
    "get_session_next_url",
    "make_handler",
    "make_token_getter",
    "oauth_error_handler",
    "oauth_logout_handler",
    "oauth1_token_setter",
    "oauth2_handle_error",
    "oauth2_token_setter",
    "response_token_setter",
    "set_session_next_url",
    "signup_handler_rest",
    "signup_handler",
    "token_delete",
    "token_getter",
    "token_session_key",
    "token_setter",
)



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Move UserIdentity to accounts."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "aaa265b0afa6"
down_revision = "bff1f190b9bd"
branch_labels = ()
depends_on = "62efc52773d4"


def upgrade():
    """Upgrade database."""
    op.execute(
        "INSERT INTO accounts_useridentity "
        "(id, method, id_user, created, updated) "
        "SELECT id, method, id_user, created, updated "
        "FROM oauthclient_useridentity;"
    )
    op.drop_table("oauthclient_useridentity")


def downgrade():
    """Downgrade database."""
    op.create_table(
        "oauthclient_useridentity",
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("method", sa.String(length=255), nullable=False),
        sa.Column("id_user", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["id_user"],
            ["accounts_user.id"],
            name=op.f("fk_oauthclient_useridentity_id_user_accounts_user"),
        ),
        sa.PrimaryKeyConstraint(
            "id", "method", name=op.f("pk_oauthclient_useridentity")
        ),
    )
    op.create_index(
        "useridentity_id_user_method",
        "oauthclient_useridentity",
        ["id_user", "method"],
        unique=True,
    )
    op.execute(
        "INSERT INTO oauthclient_useridentity "
        "(id, method, id_user, created, updated) "
        "SELECT id, method, id_user, created, updated "
        "FROM accounts_useridentity;"
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 TU Wien.
#
# Invenio-Keycloak is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helper functions for the endpoint handlers."""

import re

import jwt
from flask import current_app

from ...errors import OAuthError, OAuthKeycloakUserInfoError

AZ_09_DASHES_UNDERSCORES = r"^[A-Za-z0-9_-]+$"


def is_app_name_valid(app_name):
    """Validate app name."""
    return re.match(AZ_09_DASHES_UNDERSCORES, app_name) is not None


def _generate_config_prefix(remote):
    """Validate the app name so that it can be used in config vars."""
    app_name = remote.name
    if not is_app_name_valid(app_name):
        raise OAuthError(
            f"Invalid app name {app_name}. "
            "It should only contain letters, numbers, dashes "
            "and underscores",
            remote,
        )
    return f"OAUTHCLIENT_{app_name.upper()}"


def _format_public_key(public_key):
    """PEM-format the public key."""
    public_key = public_key.strip()

    if not public_key.startswith("-----BEGIN PUBLIC KEY-----"):
        public_key = (
            "-----BEGIN PUBLIC KEY-----\n" + public_key + "\n-----END PUBLIC KEY-----"
        )

    return public_key


def get_public_key(remote, realm_url):
    """Get the realm's public key with the ID kid from Keycloak."""
    certs_resp = remote.get(realm_url).data
    return certs_resp["public_key"]


def _get_user_info_from_token(remote, token, config_prefix):
    """Get the user information from the JWT token."""
    # try to parse the "id_token" part of Keycloak's (JWT) response
    realm_url = current_app.config[f"{config_prefix}_REALM_URL"]
    pubkey = _format_public_key(get_public_key(remote, realm_url))
    alg = jwt.get_unverified_header(token)["alg"]

    should_verify_aud = current_app.config.get(f"{config_prefix}_VERIFY_AUD", False)
    expected_aud = current_app.config.get(f"{config_prefix}_AUD", None)

    should_verify_expiration = current_app.config.get(
        f"{config_prefix}_VERIFY_EXP", False
    )

    options = {
        # check signature expiration
        "verify_exp": should_verify_expiration,
        # check the target audience
        "verify_aud": should_verify_aud and (expected_aud is not None),
    }

    return jwt.decode(
        token, key=pubkey, algorithms=[alg], audience=expected_aud, options=options
    )


def _get_user_info_from_endpoint(remote, config_prefix):
    """Get the user info from the oauth server provider."""
    url = current_app.config[f"{config_prefix}_USER_INFO_URL"]
    return remote.get(url).data


def get_user_info(remote, resp_token):
    """Get the user information from Keycloak.

    :param remote: The OAuthClient remote app
    :param resp_token: The response from the 'token' endpoint; expected to be a dict
        and to contain a JWT 'id_token'
    :returns: A tuple containing the user information extracted from the token, and
        if configured, from the UserInfo endpoint
    """
    config_prefix = _generate_config_prefix(remote)
    from_token, from_endpoint = {}, None

    try:
        from_token = _get_user_info_from_token(
            remote, resp_token["id_token"], config_prefix
        )
    except Exception as e:
        current_app.logger.exception(e)

    call_endpoint = current_app.config[f"{config_prefix}_USER_INFO_FROM_ENDPOINT"]
    if call_endpoint:
        from_endpoint = _get_user_info_from_endpoint(remote, config_prefix)

    return from_token, from_endpoint



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test handlers."""

import pytest
from flask import session, url_for
from flask_login import current_user
from flask_oauthlib.client import OAuth as FlaskOAuth
from flask_security import login_user, logout_user
from flask_security.confirmable import _security
from helpers import check_redirect_location
from werkzeug.routing import BuildError

from invenio_oauthclient import InvenioOAuthClient, current_oauthclient
from invenio_oauthclient.errors import AlreadyLinkedError
from invenio_oauthclient.handlers import (
    authorized_signup_handler,
    disconnect_handler,
    oauth_error_handler,
    signup_handler,
    token_session_key,
    token_setter,
)
from invenio_oauthclient.models import RemoteToken
from invenio_oauthclient.utils import oauth_authenticate
from invenio_oauthclient.views.client import blueprint as blueprint_client
from invenio_oauthclient.views.settings import blueprint as blueprint_settings


def test_authorized_signup_handler(remote, app, models_fixture):
    """Test authorized signup handler."""
    datastore = app.extensions["invenio-accounts"].datastore
    user = datastore.find_user(email="existing@inveniosoftware.org")
    existing_email = "existing@inveniosoftware.org"

    example_response = {"access_token": "test_access_token"}
    example_account_info = {
        "user": {
            "email": existing_email,
        },
        "external_id": "1234",
        "external_method": "test_method",
    }

    # Mock remote app's handler
    current_oauthclient.signup_handlers[remote.name] = {
        "setup": lambda token, resp: None,
        "info": lambda resp: example_account_info,
    }

    # Authenticate user
    oauth_authenticate("dev", user)

    # Mock next url
    next_url = "/test/redirect"
    session[token_session_key(remote.name) + "_next_url"] = next_url

    # Check user is redirected to next_url
    resp = authorized_signup_handler(example_response, remote)
    check_redirect_location(resp, next_url)


def test_unauthorized_signup(remote, app, models_fixture):
    """Test unauthorized redirect on signup callback handler."""
    datastore = app.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    example_response = {"access_token": "test_access_token"}
    example_account_info = {
        "user": {
            "email": existing_email,
        },
        "external_id": "1234",
        "external_method": "test_method",
    }

    # Mock remote app's handler
    current_oauthclient.signup_handlers[remote.name] = {
        "info": lambda resp: example_account_info,
    }

    _security.confirmable = True
    _security.login_without_confirmation = False
    user.confirmed_at = None
    app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name] = {}
    resp = authorized_signup_handler(example_response, remote)
    check_redirect_location(resp, lambda x: x.startswith("/login"))


def test_signup_handler(remote, app, models_fixture):
    """Test signup handler."""
    datastore = app.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    # Already authenticated
    login_user(user)
    assert current_user.is_authenticated
    resp1 = signup_handler(remote)
    check_redirect_location(resp1, "/")
    logout_user()
    assert not current_user.is_authenticated

    # No OAuth token
    resp2 = signup_handler(remote)
    check_redirect_location(resp2, "/")

    # Not coming from authorized request
    token = RemoteToken.create(user.id, "testkey", "mytoken", "mysecret")
    token_setter(remote, token, "mysecret")
    with pytest.raises(BuildError):
        signup_handler(remote)


def test_already_linked_exception(app):
    """Test error when service is already linked to another account."""

    @oauth_error_handler
    def mock_handler(resp, remote):
        raise AlreadyLinkedError(None, None)

    resp = mock_handler(None, None)
    check_redirect_location(resp, "/account/settings/linkedaccounts/")


def test_unauthorized_disconnect(app, remote):
    """Test disconnect handler when user is not authenticated."""
    resp = disconnect_handler(remote)
    check_redirect_location(resp, lambda x: x.startswith("/login/"))


def test_dummy_handler(base_app):
    """Test dummy handler."""

    # Force usage of dummy handlers
    base_app.config["OAUTHCLIENT_REMOTE_APPS"]["github"]["signup_handler"] = {}

    # Initialize InvenioOAuth
    FlaskOAuth(base_app)
    InvenioOAuthClient(base_app)
    base_app.register_blueprint(blueprint_client)
    base_app.register_blueprint(blueprint_settings)

    # Try to sign-up client
    base_app.test_client().get(
        url_for("invenio_oauthclient.signup", remote_app="github", next="/someurl/")
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

import json
from functools import partial, wraps

from flask import (
    abort,
    current_app,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
    url_for,
)
from invenio_db import db
from six.moves.urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

from ..errors import (
    AlreadyLinkedError,
    OAuthClientAlreadyAuthorized,
    OAuthClientError,
    OAuthClientMustRedirectLogin,
    OAuthClientMustRedirectSignup,
    OAuthClientTokenNotFound,
    OAuthClientTokenNotSet,
    OAuthClientUnAuthorized,
    OAuthClientUserNotRegistered,
    OAuthError,
    OAuthRejectedRequestError,
)
from ..proxies import current_oauthclient
from ..utils import create_csrf_disabled_registrationform, fill_form
from .base import (
    base_authorized_signup_handler,
    base_disconnect_handler,
    base_signup_handler,
)
from .utils import response_token_setter


def response_handler_postmessage(remote, url, payload=None):
    """Postmessage response handler."""
    return render_template("invenio_oauthclient/postmessage.html", payload=payload)


def default_response_handler(url, payload=None):
    """Default response handler."""
    scheme, netloc, path, query, fragment = urlsplit(url)
    qs = parse_qs(query)
    if payload:
        qs.update(payload)
    query = urlencode(qs)
    url = urlunsplit((scheme, netloc, path, query, fragment))
    return redirect(url)


def default_remote_response_handler(remote, url, payload=None):
    """Default response handler."""
    return default_response_handler(url, payload=payload)


def response_handler(remote, url, payload=None):
    """Handle oauthclient rest response."""
    if not remote:
        return current_oauthclient.default_response_handler(url, payload)
    return current_oauthclient.remote_app_response_handler[remote.name](
        remote, url, payload
    )


def _oauth_error_handler(remote, f, *args, **kwargs):
    """Function to handle exceptions."""
    remote_app_config = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name]
    try:
        return f(remote, *args, **kwargs)
    except OAuthClientError as e:
        current_app.logger.warning(e.message, exc_info=True)
        return response_handler(
            remote,
            remote_app_config["error_redirect_url"],
            payload=dict(message="Authorization with remote service failed.", code=400),
        )
    except OAuthRejectedRequestError:
        return response_handler(
            remote,
            remote_app_config["error_redirect_url"],
            payload=dict(message="You rejected the authentication request.", code=400),
        )
    except AlreadyLinkedError:
        msg = "External service is already linked to another account."
        return response_handler(
            remote,
            remote_app_config["error_redirect_url"],
            payload=dict(message=msg, code=400),
        )
    except OAuthClientUnAuthorized:
        return response_handler(
            remote,
            remote_app_config["error_redirect_url"],
            payload=dict(message="Unauthorized.", code=401),
        )
    except OAuthClientAlreadyAuthorized:
        return response_handler(
            remote,
            remote_app_config["authorized_redirect_url"],
            payload=dict(
                message="Successfully signed up.",
                code=200,
            ),
        )
    except OAuthClientTokenNotFound:
        return response_handler(
            remote,
            remote_app_config["error_redirect_url"],
            payload=dict(
                message="Token not found.",
                code=400,
            ),
        )
    except OAuthClientUserNotRegistered:
        abort(make_response(jsonify(message="Form validation error."), 400))
    except OAuthClientTokenNotSet:
        raise OAuthError("Could not create token for user.", remote)
    except OAuthClientMustRedirectSignup as e:
        return redirect(
            url_for(
                ".rest_signup",
                remote_app=remote.name,
            )
        )
    except OAuthClientMustRedirectLogin as e:
        return redirect(
            url_for(
                ".rest_login",
                remote_app=remote.name,
            )
        )


#
# Error handling decorators
#
def oauth_resp_remote_error_handler(f):
    """Decorator to handle exceptions."""

    @wraps(f)
    def inner(resp, remote, *args, **kwargs):
        # OAuthErrors should not happen, so they are not caught here. Hence
        # they will result in a 500 Internal Server Error which is what we
        # are interested in.
        _f = partial(f, resp)
        return _oauth_error_handler(remote, _f, *args, **kwargs)

    return inner


def oauth_remote_error_handler(f):
    """Decorator to handle exceptions."""

    @wraps(f)
    def inner(remote, *args, **kwargs):
        # OAuthErrors should not happen, so they are not caught here. Hence
        # they will result in a 500 Internal Server Error which is what we
        # are interested in.
        return _oauth_error_handler(remote, f, *args, **kwargs)

    return inner


#
# Handlers
#
@oauth_resp_remote_error_handler
def authorized_default_handler(resp, remote, *args, **kwargs):
    """Store access token in session.

    Default authorized handler.

    :param remote: The remote application.
    :param resp: The response.
    :returns: Redirect response.
    """
    remote_app_config = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name]

    response_token_setter(remote, resp)
    db.session.commit()
    return response_handler(
        remote,
        remote_app_config["authorized_redirect_url"],
    )


@oauth_resp_remote_error_handler
def authorized_signup_handler(resp, remote, *args, **kwargs):
    """Handle sign-in/up functionality.

    :param remote: The remote application.
    :param resp: The response.
    :returns: Redirect response.
    """
    remote_app_config = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name]
    next_url = base_authorized_signup_handler(resp, remote, *args, **kwargs)
    response_payload = dict(message="Successfully authorized.", code=200)
    if next_url:
        response_payload["next_url"] = next_url

    return response_handler(
        remote, remote_app_config["authorized_redirect_url"], payload=response_payload
    )


@oauth_remote_error_handler
def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    This default handler will just delete the remote account link. You may
    wish to extend this module to perform clean-up in the remote service
    before removing the link (e.g. removing install webhooks).

    :param remote: The remote application.
    :returns: Redirect response.
    """
    remote_app_config = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name]
    base_disconnect_handler(remote, *args, **kwargs)
    redirect_url = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name][
        "disconnect_redirect_url"
    ]

    return response_handler(
        remote,
        redirect_url,
        payload=dict(
            message="Successfully disconnected.",
            code=200,
        ),
    )


@oauth_remote_error_handler
def signup_handler(remote, *args, **kwargs):
    """Handle extra signup information.

    :param remote: The remote application.
    :returns: Redirect response or the template rendered.
    """
    remote_app_config = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name]
    try:
        form = create_csrf_disabled_registrationform(remote)
        data = request.form.to_dict()
        form = fill_form(form, data)
        next_url = base_signup_handler(remote, form, *args, **kwargs)
        if form.is_submitted():
            response_payload = dict(message="Successfully signed up.", code=200)
            if next_url:
                response_payload["next_url"] = next_url
            return jsonify(response_payload)
        return response_handler(
            remote,
            remote_app_config["signup_redirect_url"],
            payload=dict(
                form=form.to_dict(),
                remote=remote.name,
                app_title=remote_app_config.get("title", ""),
                app_description=remote_app_config.get("description", ""),
                app_icon=remote_app_config.get("icon", None),
            ),
        )
    except OAuthClientUnAuthorized:
        abort(401)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Advanced usage docs."""

from .ext import InvenioOAuthClient, InvenioOAuthClientREST
from .proxies import current_oauthclient

__version__ = "3.1.1"

__all__ = (
    "__version__",
    "current_oauthclient",
    "InvenioOAuthClient",
    "InvenioOAuthClientREST",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for CERN oauth remote app_rest."""

import os
from datetime import timedelta

import pytest
from flask import current_app, g, session, url_for
from flask_login import current_user
from flask_security import login_user, logout_user
from flask_security.utils import hash_password
from helpers import (
    check_response_redirect_url_args,
    get_state,
    mock_remote_get,
    mock_response,
)
from six.moves.urllib_parse import parse_qs, urlparse

from invenio_oauthclient.contrib.cern_openid import (
    OAUTHCLIENT_CERN_OPENID_SESSION_KEY,
    account_info_rest,
    disconnect_rest_handler,
    fetch_extra_data,
    get_dict_from_response,
)
from invenio_oauthclient.models import RemoteAccount

from flask_oauthlib.client import OAuthResponse  # noqa isort:skip


@pytest.fixture()
def example_cern_openid_rest(request):
    """CERN openid example data."""
    file_path = os.path.join(
        os.path.dirname(__file__), "data/cern_openid_response_content.json"
    )
    with open(file_path) as response_file:
        json_data = response_file.read()

    from jwt import encode

    token = encode(dict(name="John Doe"), key="1234")
    return (
        OAuthResponse(resp=None, content=json_data, content_type="application/json"),
        dict(
            access_token=token,
            token_type="bearer",
            expires_in=1199,
            refresh_token="test_refresh_token",
        ),
        dict(
            user=dict(
                email="john.doe@cern.ch",
                profile=dict(username="jdoe", full_name="John Doe"),
            ),
            external_id="222222",
            external_method="cern_openid",
            active=True,
        ),
    )


def test_fetch_extra_data(app_rest, example_cern_openid_rest):
    """Test extra data extraction."""
    example_response, example_token, _ = example_cern_openid_rest
    res = get_dict_from_response(example_response)

    # Check that groups were hidden as required
    extra_data = fetch_extra_data(res)

    assert "person_id" in extra_data
    assert extra_data["person_id"] == "234567"


def test_account_info_rest(app_rest, example_cern_openid_rest):
    """Test account info extraction."""
    client = app_rest.test_client()
    ioc = app_rest.extensions["oauthlib.client"]

    # Ensure remote apps have been loaded (due to before first request)
    client.get(url_for("invenio_oauthclient.rest_login", remote_app="cern_openid"))

    example_response, _, example_account_info = example_cern_openid_rest

    mock_remote_get(ioc, "cern_openid", example_response)

    assert (
        account_info_rest(ioc.remote_apps["cern_openid"], None) == example_account_info
    )


def test_account_setup(app_rest, example_cern_openid_rest, models_fixture):
    """Test account setup after login."""
    with app_rest.test_client() as c:
        ioc = app_rest.extensions["oauthlib.client"]

        # Ensure remote apps have been loaded (due to before first request)
        resp = c.get(
            url_for("invenio_oauthclient.rest_login", remote_app="cern_openid")
        )
        assert resp.status_code == 302

        example_response, example_token, example_account_info = example_cern_openid_rest

        mock_response(
            app_rest.extensions["oauthlib.client"], "cern_openid", example_token
        )
        mock_remote_get(ioc, "cern_openid", example_response)

        resp = c.get(
            url_for(
                "invenio_oauthclient.rest_authorized",
                remote_app="cern_openid",
                code="test",
                state=get_state("cern_openid"),
            )
        )
        assert resp.status_code == 302
        expected_url_args = {
            "message": "Successfully authorized.",
            "code": 200,
        }
        check_response_redirect_url_args(resp, expected_url_args)
        # 3 needs:
        # {
        #   Need(method='id', value=4),
        #   Need(method='role', value='cern_user'),
        #   Need(method='id', value='john.doe@cern.ch')
        # }
        assert len(g.identity.provides) == 3

    datastore = app_rest.extensions["invenio-accounts"].datastore
    user = datastore.find_user(email="john.doe@cern.ch")
    user.password = hash_password("1234")
    assert user
    assert user.confirmed_at

    with app_rest.test_request_context():
        resp = disconnect_rest_handler(ioc.remote_apps["cern_openid"])
        # this will delete the RemoteAccount
        assert resp.status_code >= 300

        # simulate login (account_info fetch)
        g.oauth_logged_in_with_remote = ioc.remote_apps["cern_openid"]

        login_user(user)
        # 2 needs only:
        #   missing Need(method='role', value='cern_user') that was in RemoteAccount
        assert len(g.identity.provides) == 2

        logout_user()
        assert len(g.identity.provides) == 1
        assert "cern_resource" not in session
        assert OAUTHCLIENT_CERN_OPENID_SESSION_KEY not in session

        # Login again to test the disconnect handler
        g.oauth_logged_in_with_remote = ioc.remote_apps["cern_openid"]
        login_user(user)
        assert len(g.identity.provides) == 2

        disconnect_rest_handler(ioc.remote_apps["cern_openid"])


def test_login(app_rest):
    """Test CERN login."""
    client = app_rest.test_client()

    resp = client.get(
        url_for(
            "invenio_oauthclient.rest_login", remote_app="cern_openid", next="/someurl/"
        )
    )
    assert resp.status_code == 302

    params = parse_qs(urlparse(resp.location).query)
    assert params["response_type"], ["code"]
    assert params["redirect_uri"]
    assert params["client_id"]
    assert params["state"]


def test_authorized_reject(app_rest):
    """Test a rejected request."""
    with app_rest.test_client() as c:
        c.get(url_for("invenio_oauthclient.rest_login", remote_app="cern_openid"))
        resp = c.get(
            url_for(
                "invenio_oauthclient.rest_authorized",
                remote_app="cern_openid",
                error="access_denied",
                error_description="User denied access",
                state=get_state("cern_openid"),
            )
        )
        assert resp.status_code in (301, 302)
        expected_url_args = {
            "message": "You rejected the authentication request.",
            "code": 400,
        }
        check_response_redirect_url_args(resp, expected_url_args)


def test_account_info_not_allowed_account(app_rest, example_cern_openid_rest):
    """Test account info extraction."""
    client = app_rest.test_client()

    app_rest.config["OAUTHCLIENT_CERN_OPENID_ALLOWED_ROLES"] = ["another cern role"]
    ioc = app_rest.extensions["oauthlib.client"]

    # Ensure remote apps have been loaded (due to before first request)
    client.get(url_for("invenio_oauthclient.rest_login", remote_app="cern_openid"))

    example_response, _, example_account_info = example_cern_openid_rest

    mock_remote_get(ioc, "cern_openid", example_response)
    resp = account_info_rest(ioc.remote_apps["cern_openid"], None)
    assert g.oauth_logged_in_with_remote == ioc.remote_apps["cern_openid"]

    assert resp.status_code == 302
    expected_url_args = {
        "message": "CERN account not allowed.",
        "code": 400,
    }
    check_response_redirect_url_args(resp, expected_url_args)


def test_identity_changed(app_rest, example_cern_openid_rest, models_fixture):
    def _init():
        ioc = app_rest.extensions["oauthlib.client"]

        # setup the user account via cern_openid
        with app_rest.test_client() as c:
            # Ensure remote apps have been loaded (due to before first request)
            resp = c.get(
                url_for("invenio_oauthclient.rest_login", remote_app="cern_openid")
            )
            assert resp.status_code == 302

            (
                example_response,
                example_token,
                example_account_info,
            ) = example_cern_openid_rest

            mock_response(
                app_rest.extensions["oauthlib.client"], "cern_openid", example_token
            )
            mock_remote_get(ioc, "cern_openid", example_response)

            resp = c.get(
                url_for(
                    "invenio_oauthclient.rest_authorized",
                    remote_app="cern_openid",
                    code="test",
                    state=get_state("cern_openid"),
                )
            )
            assert resp.status_code == 302
            expected_url_args = {
                "message": "Successfully authorized.",
                "code": 200,
            }
            check_response_redirect_url_args(resp, expected_url_args)

            assert len(g.identity.provides) == 3

    def _test_with_token(user, remote_account):
        with app_rest.test_request_context():
            # mark user as logged in via token
            user.login_via_oauth2 = True
            # check if the initial roles are there
            login_user(user)

            assert current_user.login_via_oauth2

            # 3 needs:
            # {
            #   Need(method='id', value=4),
            #   Need(method='role', value='cern_user'),
            #   Need(method='id', value='john.doe@cern.ch')
            # }
            assert len(g.identity.provides) == 3
            logout_user()

            # remove the cern roles
            remote_account.extra_data.update(roles=[])

            # login the user again
            login_user(user)

            # 2 needs:
            # {
            #   Need(method='id', value=4),
            #   Need(method='role', value='cern_user'),
            # }
            assert len(g.identity.provides) == 2
            logout_user()

    def _test_without_token(user, remote_account):
        user.login_via_oauth2 = False

        current_app.config["OAUTHCLIENT_CERN_OPENID_REFRESH_TIMEDELTA"] = False
        login_user(user)

        # check that the roles are not refreshed from provider
        assert len(g.identity.provides) == 2
        logout_user()

        current_app.config["OAUTHCLIENT_CERN_OPENID_REFRESH_TIMEDELTA"] = timedelta(
            microseconds=1
        )
        login_user(user)

        # check if roles refreshed from the provider
        assert len(g.identity.provides) == 3

    _init()

    datastore = app_rest.extensions["invenio-accounts"].datastore
    user = datastore.find_user(email="john.doe@cern.ch")
    assert user

    client_id = current_app.config["CERN_APP_OPENID_CREDENTIALS"]["consumer_key"]
    # make sure the roles are cleaned
    remote_account = RemoteAccount.get(user_id=user.get_id(), client_id=client_id)

    _test_with_token(user, remote_account)
    _test_without_token(user, remote_account)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test example app."""

import os
import signal
import subprocess
import time
from contextlib import contextmanager

import pytest


@contextmanager
def _create_example_app(app_name):
    """Example app fixture."""
    current_dir = os.getcwd()
    # go to example directory
    project_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    exampleappdir = os.path.join(project_dir, "examples")
    os.chdir(exampleappdir)
    # setup example
    cmd = "FLASK_APP={0} ./app-setup.sh".format(app_name)
    exit_status = subprocess.call(cmd, shell=True)
    assert exit_status == 0
    # Starting example web app
    cmd = "FLASK_APP={0} flask run --debugger -p 5000".format(app_name)
    webapp = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, preexec_fn=os.setsid, shell=True
    )
    time.sleep(5)
    # return webapp
    yield webapp
    # stop server
    os.killpg(webapp.pid, signal.SIGTERM)
    # tear down example app
    cmd = "FLASK_APP={0} ./app-teardown.sh".format(app_name)
    subprocess.call(cmd, shell=True)
    # return to the original directory
    os.chdir(current_dir)
    time.sleep(2)


@pytest.mark.parametrize("service", ["orcid", "github", "globus"])
def test_app(service):
    """Test example app for given service."""
    with _create_example_app("{0}_app.py".format(service)):
        cmd = "curl http://localhost:5000/{0}".format(service)
        output = subprocess.check_output(cmd, shell=True).decode("utf-8")
        assert "Redirect" in output


@pytest.mark.parametrize("service", ["orcid", "github", "globus"])
def test_app_rest(service):
    """Test example app for given service."""
    with _create_example_app("{0}_app_rest.py".format(service)):
        cmd = "curl http://localhost:5000/{0}".format(service)
        output = subprocess.check_output(cmd, shell=True).decode("utf-8")
        assert "Redirect" in output



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pre-configured remote application for enabling sign in/up with OpenAIRE.

1. Edit your configuration and add:

   .. code-block:: python

       from invenio_oauthclient.contrib import openaire_aai

       OAUTHCLIENT_REMOTE_APPS = dict(
           openaire_aai=openaire_aai.REMOTE_APP,
       )

       OPENAIRE_APP_CREDENTIALS = dict(
           consumer_key="changeme",
           consumer_secret="changeme",
       )

  In case you want use sandbox ``openaire_aai.REMOTE_SANDBOX_APP`` instead
  of ``openaire_aai.REMOTE_APP``.

2. Register a new application with OpenAIRE, by sending an email to
   <aai@openaire.eu>, with the following information:

   - Client ID for your application (e.g. ``my-app``).
   - One or more *Redirect URI*s pointing to
    ``https://<host>/oauth/authorized/openaire_aai/`` (e.g.
    ``https://localhost:5000/oauth/authorized/openaire_aai/``).
   - User claim scopes ``openid profile email orcid``.
   - One or more of the OpenID Connect/OAuth2 grant types: Authorization Code,
    Token Exchange, Device Code.

3. Once you have your *Client ID* and *Client Secret* add them to your instance
   configuration (e.g. ``invenio.cfg``):

   .. code-block:: python

        OPENAIRE_APP_CREDENTIALS = dict(
           consumer_key="<CLIENT ID>",
           consumer_secret="<CLIENT SECRET>",
        )

4. Now go to ``https://<host>/oauth/login/openaire_aai/`` (e.g.
   https://localhost:5000/oauth/login/openaire_aai/)

5. Also, you should see OpenAIRE listed under Linked accounts:
   https://localhost:5000/account/settings/linkedaccounts/

By default the OpenAIRE module will try first look if a link already exists
between a OpenAIRE account and a user. If no link is found, the user is asked
to provide an email address to sign-up.

In templates you can add a sign in/up link:

.. code-block:: jinja

    <a href="{{url_for('invenio_oauthclient.login',
        remote_app='openaire_aai')}}">
      Sign in with OpenAIRE
    </a>

"""

import jwt
from flask import current_app, redirect, url_for
from flask_login import current_user
from invenio_db import db

from invenio_oauthclient import current_oauthclient
from invenio_oauthclient.contrib.keycloak import KeycloakSettingsHelper
from invenio_oauthclient.contrib.settings import OAuthSettingsHelper
from invenio_oauthclient.handlers.rest import response_handler
from invenio_oauthclient.handlers.utils import require_more_than_one_external_account
from invenio_oauthclient.models import RemoteAccount
from invenio_oauthclient.utils import oauth_link_external_id, oauth_unlink_external_id


class OpenAIREAuthSettingsHelper(OAuthSettingsHelper):
    """Default configuration for OpenAIRE OAuth provider."""

    def __init__(
        self,
        title=None,
        description=None,
        base_url=None,
        app_key=None,
        precedence_mask=None,
        signup_options=None,
    ):
        """Constructor."""
        base_url = base_url or "https://aai.openaire.eu"
        super().__init__(
            title or "OpenAIRE",
            description or "Open Science Services.",
            base_url,
            app_key or "OPENAIRE_APP_CREDENTIALS",
            request_token_params={"scope": "openid profile email orcid"},
            access_token_url=f"{base_url}/oidc/token",
            authorize_url=f"{base_url}/oidc/authorize",
            content_type="application/json",
            precedence_mask=precedence_mask or {"email": True},
            signup_options=signup_options,
        )

        self._handlers = dict(
            authorized_handler="invenio_oauthclient.handlers:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.openaire_aai:disconnect_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.openaire_aai:account_info",
                info_serializer="invenio_oauthclient.contrib.openaire_aai:account_info_serializer",
                setup="invenio_oauthclient.contrib.openaire_aai:account_setup",
                view="invenio_oauthclient.handlers:signup_handler",
            ),
        )

        self._rest_handlers = dict(
            authorized_handler="invenio_oauthclient.handlers.rest:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.openaire_aai:disconnect_rest_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.openaire_aai:account_info",
                info_serializer="invenio_oauthclient.contrib.openaire_aai:account_info_serializer",
                setup="invenio_oauthclient.contrib.openaire_aai:account_setup",
                view="invenio_oauthclient.handlers.rest:signup_handler",
            ),
            response_handler="invenio_oauthclient.handlers.rest:default_remote_response_handler",
            authorized_redirect_url="/",
            disconnect_redirect_url="/",
            signup_redirect_url="/",
            error_redirect_url="/",
        )

    @property
    def user_info_url(self):
        """User info URL."""
        return f"{self.base_url}/oidc/userinfo"

    def get_handlers(self):
        """Return OpenAIRE auth handlers."""
        return self._handlers

    def get_rest_handlers(self):
        """Return OpenAIRE auth REST handlers."""
        return self._rest_handlers


# Production
_openaire_aai_app = OpenAIREAuthSettingsHelper()

BASE_APP = _openaire_aai_app.base_app
REMOTE_APP = _openaire_aai_app.remote_app
"""OpenAIRE Remote Application."""

REMOTE_REST_APP = _openaire_aai_app.remote_rest_app
"""OpenAIRE Remote REST Application."""

# Sandbox
_openaire_aai_sandbox_app = KeycloakSettingsHelper(
    title="OpenAIRE",
    description="Open Science Services.",
    base_url="https://beta.aai.openaire.eu",
    realm="openaire",
    scopes="openid profile email eduperson_entitlement orcid",
    app_key="OPENAIRE_APP_CREDENTIALS",
)
REMOTE_SANDBOX_APP = _openaire_aai_sandbox_app.remote_app
"""OpenAIRE Sandbox Remote Application."""

REMOTE_SANDBOX_REST_APP = _openaire_aai_sandbox_app.remote_rest_app
"""OpenAIRE Sandbox Remote REST Application."""

OAUTHCLIENT_OPENAIRE_AAI_JWT_DECODE_PARAMS = dict(
    options=dict(
        verify_signature=False,
        verify_aud=False,
    ),
    algorithms=[
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512",
    ],
)
"""OpenAIRE AAI JWT decoding parameters."""


def account_info_serializer(remote, resp, user_info, **kwargs):
    """Serialize the account info response object.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :param user_info: The response of the `user info` endpoint.
    :returns: A dictionary with serialized user information.
    """
    return {
        "external_id": user_info["sub"],
        "external_method": remote.name,
        "user": {
            "profile": {
                "full_name": user_info.get("name"),
            },
            "email": user_info.get("email"),
        },
    }


def account_info(remote, resp):
    """Retrieve remote account information used to find local user.

    It returns a dictionary with the following structure:

    .. code-block:: python

        {
            'user': {
                'profile': {
                    'full_name': 'Full Name',
                },
                'email': '<openaire-email>',
            },
            'external_id': '<openaire-unique-identifier>',
            'external_method': 'openaire_aai',
        }

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: A dictionary with the user information.
    """
    user_info_url = f"{remote.base_url}/oidc/userinfo"
    user_info = remote.get(user_info_url).data

    handlers = current_oauthclient.signup_handlers[remote.name]
    # `remote` param automatically injected via `make_handler` helper
    return handlers["info_serializer"](resp, user_info)


@require_more_than_one_external_account
def _disconnect(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    """
    if not current_user.is_authenticated:
        return current_app.login_manager.unauthorized()

    remote_account = RemoteAccount.get(
        user_id=current_user.get_id(), client_id=remote.consumer_key
    )
    external_ids = [
        i.id for i in current_user.external_identifiers if i.method == remote.name
    ]

    if external_ids:
        oauth_unlink_external_id(dict(id=external_ids[0], method=remote.name))
    if remote_account:
        with db.session.begin_nested():
            remote_account.delete()


def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    """
    _disconnect(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


def disconnect_rest_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    """
    _disconnect(remote, *args, **kwargs)
    redirect_url = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name][
        "disconnect_redirect_url"
    ]
    return response_handler(remote, redirect_url)


def account_setup(remote, token, resp):
    """Perform additional setup after user have been logged in.

    :param remote: The remote application.
    :param token: The token value.
    :param resp: The response.
    """
    user_info = jwt.decode(
        resp["id_token"],
        **current_app.config.get(
            "OAUTHCLIENT_OPENAIRE_AAI_JWT_DECODE_PARAMS",
            OAUTHCLIENT_OPENAIRE_AAI_JWT_DECODE_PARAMS,
        ),
    )
    openaire_id = user_info["sub"]
    with db.session.begin_nested():
        user = token.remote_account.user

        # Set extra data so that we mark that the setup is done.
        token.remote_account.extra_data = {"id": openaire_id}

        # Create user <-> external id link.
        oauth_link_external_id(
            user,
            {"id": openaire_id, "method": remote.name},
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C)      2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Configuration variables for defining remote applications.

================================ ==============================================
`OAUTHCLIENT_REMOTE_APPS`        Dictionary of remote applications. See example
                                 below. **Default:** ``{}``.
`OAUTHCLIENT_SESSION_KEY_PREFIX` Prefix for the session key used to store the
                                 an access token. **Default:** ``oauth_token``.
`OAUTHCLIENT_STATE_EXPIRES`      Number of seconds after which the state token
                                 expires. Defaults to 300 seconds.
`OAUTHCLIENT_REMOTE_APP`         Replaces the default remote application class.
================================ ==============================================

Each remote application must be defined in the ``OAUTHCLIENT_REMOTE_APPS``
dictionary, where the keys are the application names and the values the
configuration parameters for the application.

.. code-block:: python

    OAUTHCLIENT_REMOTE_APPS = dict(
        myapp=dict(
            # configuration values for myapp ...
        ),
    )

The application name is used in the login, authorized, sign-up and disconnect
endpoints:

- Login endpoint: ``/oauth/login/<REMOTE APP>/``.
- Authorized endpoint: ``/oauth/authorized/<REMOTE APP>/``.
- Disconnect endpoint: ``/oauth/disconnect/<REMOTE APP>/``.
- Sign up endpoint: ``/oauth/login/<REMOTE APP>/``.


Remote application
^^^^^^^^^^^^^^^^^^
Configuration of a single remote application is a dictionary with the following
keys:

- ``title`` - Title of remote application. Displayed to end-users under Account
  > Linked accounts.
- ``description`` - Short description of remote application. Displayed to
  end-users under Account > Linked accounts.
- ``icon`` - CSS class for icon of service (e.g. ``fa fa-github`` for using the
  Font-Awesome GitHub icon). Displayed to end-users.
- ``params`` - Flask-OAuthlib remote application parameters..
- ``authorized_handler`` - Import path to authorized callback handler.
- ``disconnect_handler`` - Import path to disconnect callback handler.
- ``signup_handler`` - A dictionary of import path to sign up callback handler.
- ``precedence_mask`` - A mask determining which user info values should
  override user input during sign-up.

.. code-block:: python

    OAUTHCLIENT_REMOTE_APPS = dict(
        myapp=dict(
            title='...',
            description='...',
            icon='...',
            authorized_handler="...",
            disconnect_handler="...",
            signup_handler=dict(
                info="...",
                setup="...",
                view="...",
                groups="...",
                groups_serializer="...",
            ),
            precedence_mask=dict(
                email=True
            ),
            signup_options=dict(
                auto_confirm=True,
                send_register_msg=False,
            ),
            params=dict(...),
            )
        )
    )

Note on the ``precedence_mask``:

This mask is used during sign-up of new users via external OAuth
providers, to determine for which of the new user's properties the
`user_info` given by the OAuth provider must take precedence over
any user input.

Properties marked with `False` in the precedence mask will be used
as specified by the user.
Any properties marked with `True` (or not appearing) in the precedence
mask will be taken from the OAuth service's `user_info` dictionary,
overriding any potential user input from registration forms.
If a property is missing from the `user_info` dictionary, the
configured value in the precedence mask is irrelevant.

For instance, if the following user info were given by the OAuth
remote app during signup, and the precedence mask configured as
follows:

.. code-block:: python

    # user_info from the oauth remote app
    {
        "email": "user@inveniosoftware.org",
        "password": "somepassword",
        "profile": {
            "username": "test-user",
            "full_name": "Test User",
        }
    }

    # precedence_mask
    {
        "email": True,
        "profile": {
            "username": True,
            "full_name": False,
        }
    }

Then, the values for `email` and `profile.username` from the
`user_info` would be overriding any user input from the registration
form.
All other values would be used as provided by the user in the form.

WARNING: Allowing users to specify their email address arbitrarily
(`precedence_mask["email"] = False`) during sign-up may have severe
security implications, as the linking of external accounts with
accounts in Invenio is done by matching email addresses!

Note on the ``signup_options``:

This parameter accepts a dictionary with `auto_confirm` and `send_register_msg`
boolean values.

When `auto_confirm` is set to `True` (default), the user, after the first login, will
be automatically confirmed and will not receive an e-mail with a link to confirm
the account. This is the common behaviour normally expected when authentication
happens via an external provider. The only exception is the ORCID contribution,
given that the e-mail is input by the user (and therefor should be confirmed).

The `send_register_msg` allows to set if the welcome e-mail, after the first login,
should be sent. Normally, it should be disabled when using OAuth. `False` by
default.


Remote REST application
^^^^^^^^^^^^^^^^^^^^^^^
Configuration of a single remote REST application is a dictionary with the
same keys as Remote application in addition to:

- ``response_handler`` - Import path to response callback handler.
- ``authorized_redirect_url`` - URL path to redirect your SPA when login was
  successfull.
- ``disconnect_redirect_url`` - URL path to redirect your SPA after logging
  out.
- ``signup_redirect_url`` - URL path to redirect your SPA to sign up a user.
- ``error_redirect_url`` - URL path to redirect your SPA when an error occurred


.. code-block:: python

    OAUTHCLIENT_REMOTE_APPS = dict(
        myapp=dict(
            title='...',
            description='...',
            icon='...',
            authorized_handler="...",
            disconnect_handler="...",
            signup_handler=dict(
                info="...",
                info_serializer="...",
                setup="...",
                view="...",
                groups="...",
                groups_serializer="...",
            ),
            response_handler=("..."),
            authorized_redirect_url="...",
            disconnect_redirect_url="...",
            signup_redirect_url="...",
            error_redirect_url="...",
            precedence_mask=dict(...),
            params=dict(...),
            )
        )
    )


Flask-OAuthlib parameters
^^^^^^^^^^^^^^^^^^^^^^^^^
The Flask-OAuthlib parameters defines the remote application OAuth endpoints as
well as the client id and secret. Full description of these parameters are
given in the `Flask-OAuthlib documentation
<https://flask-oauthlib.readthedocs.io/en/latest/client.html>`_.

Normally you will have to browse the remote application's API documentation to
find which URLs and scopes to use.

Below is an example for GitHub:

.. code-block:: python

    OAUTHCLIENT_REMOTE_APPS = dict(
        github=dict(
            # ...
            params=dict(
                request_token_params={'scope': 'user:email'},
                base_url='https://api.github.com/',
                request_token_url=None,
                access_token_url="https://github.com/login/oauth/access_token",
                access_token_method='POST',
                authorize_url="https://github.com/login/oauth/authorize",
                app_key="GITHUB_APP_CREDENTIALS",
            )
        )
    )

    GITHUB_APP_CREDENTIALS=dict(
        consumer_key="changeme"
        consumer_secret="changeme"
    )

The ``app_key`` parameter allows you to put your sensitive client id and secret
in your instance configuration (``var/invenio.base-instance/invenio.cfg``).

Handlers
^^^^^^^^
Handlers allow customizing oauthclient endpoints for each remote
application:

- Authorized endpoint: ``/oauth/authorized/<REMOTE APP>/``.
- Disconnect endpoint: ``/oauth/disconnect/<REMOTE APP>/``.
- Sign up endpoint: ``/oauth/login/<REMOTE APP>/``.

By default only authorized and disconnect handlers are required, and Invenio
provide default implementation that stores the access token in the user session
as well as to the database if the user is authenticated:


.. code-block:: python

    OAUTHCLIENT_REMOTE_APPS = dict(
        myapp=dict(
            # ...
            authorized_handler="invenio_oauthclient.handlers"
                       ":authorized_default_handler",
            disconnect_handler="invenio_oauthclient.handlers"
                       ":disconnect_handler",
            )
            # ...
        )
    )

If you want to provide sign in/up functionality using oauthclient, Invenio
comes with a default handler that will try to find a matching local user for
a given authorize request.

.. code-block:: python

    OAUTHCLIENT_REMOTE_APPS = dict(
        orcid=dict(
            # ...
            authorized_handler="invenio_oauthclient.handlers"
                       ":authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.handlers"
                       ":disconnect_handler",
            )
            signup_handler=dict(
                info="invenio_oauthclient.contrib.orcid:account_info",
                info_serializer="invenio_oauthclient.contrib.orcid:account_info_serializer",
                setup="invenio_oauthclient.contrib.orcid:account_setup",
                view="invenio_oauthclient.handlers:signup_handler",
                groups="invenio_oauthclient.handlers:signup_handler",
                groups_serializer="invenio_oauthclient.handlers:groups_serializer",
            ),
            # ...
        )
    )

Custom remote application
^^^^^^^^^^^^^^^^^^^^^^^^^

Some OAuth services require a specific handling of OAuth requests. If the
standard flask-oauthlib.client.OAuthRemoteApp does not support it, it is
possible to replace the standard OAuthRemoteApp for all remote application
by referring to the custom class with the configuration variable
``OAUTHCLIENT_REMOTE_APP`` or for only one remote application by
setting ``remote_app`` in your remote application configuration.

.. code-block:: python

    class CustomOAuthRemoteApp(OAuthRemoteApp):
        pass

    app.config.update(
        OAUTHCLIENT_REMOTE_APP=
            'myproject.mymodule:CustomOAuthRemoteApp'
    )

    # OR

    app.config.update(
        OAUTHCLIENT_REMOTE_APPS=dict(
            custom_app=dict(
                # ...
                remote_app=
                    'myproject.mymodule:CustomOAuthRemoteApp'
            )
        )
    )

"""

from invenio_oauthclient.utils import _create_registrationform

from .views.client import auto_redirect_login

OAUTHCLIENT_REMOTE_APPS = {}
"""Configuration of remote applications."""

OAUTHCLIENT_SESSION_KEY_PREFIX = "oauth_token"
"""Session key prefix used when storing the access token for a remote app."""

OAUTHCLIENT_STATE_EXPIRES = 300
"""Number of seconds after which the state token expires."""

OAUTHCLIENT_STATE_ENABLED = True
"""Internal variable used to disable state validation during tests."""

OAUTHCLIENT_SIGNUP_FORM = _create_registrationform
"""Function called to render the sign up form after authorization succeeded."""

OAUTHCLIENT_SIGNUP_TEMPLATE = "invenio_oauthclient/signup.html"
"""Template for the signup page."""

OAUTHCLIENT_REST_REMOTE_APPS = {}
"""Configuration of remote rest applications."""

OAUTHCLIENT_REST_DEFAULT_ERROR_REDIRECT_URL = "/"
"""Configuration of default error redirect URL."""

OAUTHCLIENT_REST_DEFAULT_RESPONSE_HANDLER = None
"""Default REST response handler."""

OAUTHCLIENT_AUTO_REDIRECT_TO_EXTERNAL_LOGIN = False
"""Redirect to the only external login service under specific conditions.

If this option is enabled and there is exactly one external authentication
service enabled (i.e. one OAuthClient remote app is configured, and local
login is disabled), the login view function will automatically redirect to
this external authentication service.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pre-configured remote application for enabling sign in/up with ORCID.

1. Edit your configuration and add:

   .. code-block:: python

       from invenio_oauthclient.contrib import orcid

       OAUTHCLIENT_REMOTE_APPS = dict(
           orcid=orcid.REMOTE_APP,
       )

       ORCID_APP_CREDENTIALS = dict(
           consumer_key="changeme",
           consumer_secret="changeme",
       )

  Note, if you want to use the ORCID Member API, use
  ``orcid.REMOTE_MEMBER_APP`` instead of ``orcid.REMOTE_APP``.

  In case you want use sandbox:
  To use the ORCID Public API sandbox, use ``orcid.REMOTE_SANDBOX_APP``
  instead of ``orcid.REMOTE_APP``.
  To use the ORCID Member API sandbox, use ``orcid.REMOTE_SANDBOX_MEMBER_APP``.

2. Register a new application with ORCID. When registering the
   application ensure that the *Redirect URI* points to:
   ``CFG_SITE_URL/oauth/authorized/orcid/`` (note, ORCID does not
   allow localhost to be used, thus testing on development machines is
   somewhat complicated by this).


3. Grab the *Client ID* and *Client Secret* after registering the application
   and add them to your instance configuration (``invenio.cfg``):

   .. code-block:: python

       ORCID_APP_CREDENTIALS = dict(
           consumer_key="<CLIENT ID>",
           consumer_secret="<CLIENT SECRET>",
       )

4. Now go to ``CFG_SITE_URL/oauth/login/orcid/`` (e.g.
   http://localhost:4000/oauth/login/orcid/)

5. Also, you should see ORCID listed under Linked accounts:
   http://localhost:4000/account/settings/linkedaccounts/

By default the ORCID module will try first look if a link already exists
between a ORCID account and a user. If no link is found, the user is asked
to provide an email address to sign-up.

In templates you can add a sign in/up link:

.. code-block:: jinja

    <a href="{{url_for('invenio_oauthclient.login', remote_app='orcid')}}">
      Sign in with ORCID
    </a>


For more details you can play with a :doc:`working example <examplesapp>`.

"""

from flask import current_app, redirect, url_for
from flask_login import current_user
from invenio_db import db

from invenio_oauthclient import current_oauthclient
from invenio_oauthclient.contrib.settings import OAuthSettingsHelper
from invenio_oauthclient.handlers.rest import response_handler
from invenio_oauthclient.handlers.utils import require_more_than_one_external_account
from invenio_oauthclient.models import RemoteAccount
from invenio_oauthclient.utils import oauth_link_external_id, oauth_unlink_external_id


class ORCIDOAuthSettingsHelper(OAuthSettingsHelper):
    """Default configuration for ORCID OAuth provider."""

    def __init__(
        self,
        title=None,
        description=None,
        base_url=None,
        app_key=None,
        access_token_url=None,
        authorize_url=None,
        precedence_mask=None,
        signup_options=None,
    ):
        """Constructor."""
        access_token_url = access_token_url or "https://orcid.org/oauth/token"
        authorize_url = authorize_url or "https://orcid.org/oauth/authorize"
        precedence_mask = precedence_mask or {
            "email": False,
        }
        signup_options = signup_options or {
            "auto_confirm": False,
            "send_register_msg": True,
        }

        super().__init__(
            title or "ORCID",
            description or "Connecting Research and Researchers.",
            base_url or "https://pub.orcid.org/v1.2/",
            app_key or "ORCID_APP_CREDENTIALS",
            request_token_params={"scope": "/authenticate", "show_login": "true"},
            access_token_url=access_token_url,
            authorize_url=authorize_url,
            content_type="application/json",
            precedence_mask=precedence_mask,
            signup_options=signup_options,
        )

        self._handlers = dict(
            authorized_handler="invenio_oauthclient.handlers:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.orcid:disconnect_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.orcid:account_info",
                info_serializer="invenio_oauthclient.contrib.orcid:account_info_serializer",
                setup="invenio_oauthclient.contrib.orcid:account_setup",
                view="invenio_oauthclient.handlers:signup_handler",
            ),
        )

        self._rest_handlers = dict(
            authorized_handler="invenio_oauthclient.handlers.rest:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.orcid:disconnect_rest_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.orcid:account_info",
                info_serializer="invenio_oauthclient.contrib.orcid:account_info_serializer",
                setup="invenio_oauthclient.contrib.orcid:account_setup",
                view="invenio_oauthclient.handlers.rest:signup_handler",
            ),
            response_handler="invenio_oauthclient.handlers.rest:default_remote_response_handler",
            authorized_redirect_url="/",
            disconnect_redirect_url="/",
            signup_redirect_url="/",
            error_redirect_url="/",
        )

    def get_handlers(self):
        """Return ORCID auth handlers."""
        return self._handlers

    def get_rest_handlers(self):
        """Return ORCID auth REST handlers."""
        return self._rest_handlers


_orcid_app = ORCIDOAuthSettingsHelper()

BASE_APP = _orcid_app.base_app
REMOTE_APP = _orcid_app.remote_app
"""ORCID Remote Application."""

REMOTE_REST_APP = _orcid_app.remote_rest_app
"""ORCID Remote REST Application."""

_orcid_member_app = ORCIDOAuthSettingsHelper(
    base_url="https://api.orcid.org/", access_token_url="https://orcid.org/oauth/token"
)
REMOTE_MEMBER_APP = _orcid_member_app.remote_app
"""ORCID Remote Application with member API."""


# SANDBOX

_orcid_sandbox_app = ORCIDOAuthSettingsHelper(
    base_url="https://pub.sandbox.orcid.org/",
    access_token_url="https://sandbox.orcid.org/oauth/token",
    authorize_url="https://sandbox.orcid.org/oauth/authorize#show_login",
)
REMOTE_SANDBOX_APP = _orcid_sandbox_app.remote_app
"""ORCID Sandbox Remote Application with public API."""

REMOTE_SANDBOX_REST_APP = _orcid_sandbox_app.remote_rest_app
"""ORCID Sandbox Remote Application with public API."""

_orcid_sandbox_member_app = ORCIDOAuthSettingsHelper(
    base_url="https://api.sandbox.orcid.org/",
    access_token_url="https://sandbox.orcid.org/oauth/token",
    authorize_url="https://sandbox.orcid.org/oauth/authorize#show_login",
)
REMOTE_SANDBOX_MEMBER_APP = _orcid_sandbox_member_app.remote_app
"""ORCID sandbox member API."""


def account_info_serializer(remote, resp, **kwargs):
    """Serialize the account info response object.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: A dictionary with serialized user information.
    """
    return {
        "external_id": resp.get("orcid"),
        "external_method": remote.name,
        "user": {
            "profile": {
                "full_name": resp.get("name"),
            },
        },
    }


def account_info(remote, resp):
    """Retrieve remote account information used to find local user.

    It returns a dictionary with the following structure:

    .. code-block:: python

        {
            'user': {
                'profile': {
                    'full_name': 'Full Name',
                },
            },
            'external_id': 'orcid-unique-identifier',
            'external_method': 'orcid',
        }

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: A dictionary with the user information.
    """
    handlers = current_oauthclient.signup_handlers[remote.name]
    # `remote` param automatically injected via `make_handler` helper
    return handlers["info_serializer"](resp)


@require_more_than_one_external_account
def _disconnect(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    """
    if not current_user.is_authenticated:
        return current_app.login_manager.unauthorized()

    account = RemoteAccount.get(
        user_id=current_user.get_id(), client_id=remote.consumer_key
    )
    orcid = account.extra_data.get("orcid")

    if orcid:
        oauth_unlink_external_id({"id": orcid, "method": "orcid"})
    if account:
        with db.session.begin_nested():
            account.delete()


def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    """
    _disconnect(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


def disconnect_rest_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    """
    _disconnect(remote, *args, **kwargs)
    redirect_url = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name][
        "disconnect_redirect_url"
    ]
    return response_handler(remote, redirect_url)


def account_setup(remote, token, resp):
    """Perform additional setup after user have been logged in.

    :param remote: The remote application.
    :param token: The token value.
    :param resp: The response.
    """
    with db.session.begin_nested():
        # Retrieve ORCID from response.
        orcid = resp.get("orcid")
        full_name = resp.get("name")

        # Set ORCID in extra_data.
        token.remote_account.extra_data = {
            "orcid": orcid,
            "full_name": full_name,
        }

        user = token.remote_account.user

        # Create user <-> external id link.
        oauth_link_external_id(user, {"id": orcid, "method": "orcid"})



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test handlers."""

import pytest
from flask import session, url_for
from flask_login import current_user
from flask_oauthlib.client import OAuth as FlaskOAuth
from flask_security import login_user, logout_user
from flask_security.confirmable import _security
from helpers import check_response_redirect_url_args
from invenio_accounts.models import Role
from werkzeug.routing import BuildError

from invenio_oauthclient import InvenioOAuthClientREST, current_oauthclient
from invenio_oauthclient.errors import AlreadyLinkedError
from invenio_oauthclient.handlers import token_session_key, token_setter
from invenio_oauthclient.handlers.rest import (
    authorized_signup_handler,
    disconnect_handler,
    oauth_resp_remote_error_handler,
    response_handler_postmessage,
    signup_handler,
)
from invenio_oauthclient.models import RemoteToken
from invenio_oauthclient.utils import oauth_authenticate
from invenio_oauthclient.views.client import rest_blueprint

REMOTE_APPS = ["github", "orcid", "globus"]


@pytest.fixture(scope="function")
def remote(request, app_rest):
    """Fixture to return a remote app by name."""
    oauth = current_oauthclient.oauth
    return oauth.remote_apps[request.param]


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_authorized_signup_handler(remote, app_rest, models_fixture):
    """Test authorized signup handler."""
    datastore = app_rest.extensions["invenio-accounts"].datastore
    user = datastore.find_user(email="existing@inveniosoftware.org")

    example_response = {"access_token": "test_access_token"}

    # Mock remote app's handler
    current_oauthclient.signup_handlers[remote.name] = {
        "setup": lambda token, resp: None
    }

    # Authenticate user
    oauth_authenticate("dev", user)

    # Mock next url
    next_url = "/test/redirect"
    session[token_session_key(remote.name) + "_next_url"] = next_url

    # Check user is redirected to next_url
    expected_url_args = {
        "message": "Successfully authorized.",
        "code": 200,
        "next_url": next_url,
    }
    resp = authorized_signup_handler(example_response, remote)
    check_response_redirect_url_args(resp, expected_url_args)


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_group_handler(remote, app_rest, models_fixture):
    """Test group handler."""
    datastore = app_rest.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)
    example_group = [
        {
            "id": "rdm-developers",
            "name": "rdm-developers",
            "description": "People contributing to RDM.",
        }
    ]

    example_response = {"access_token": "test_access_token"}
    example_account_info = {
        "user": {
            "email": existing_email,
        },
        "external_id": "1234",
        "external_method": "test_method",
    }

    # Mock remote app's handler
    current_oauthclient.signup_handlers[remote.name] = {
        "info": lambda resp: example_account_info,
        "groups": lambda resp: example_group,
    }

    _security.confirmable = True
    _security.login_without_confirmation = False
    user.confirmed_at = None

    authorized_signup_handler(example_response, remote)

    # Assert that the group handler works correctly
    roles = Role.query.all()
    assert 1 == len(roles)
    assert roles[0].id == example_group[0]["id"]
    assert roles[0].name == example_group[0]["name"]
    assert roles[0].description == example_group[0]["description"]


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_unauthorized_signup(remote, app_rest, models_fixture):
    """Test unauthorized redirect on signup callback handler."""
    datastore = app_rest.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    example_response = {"access_token": "test_access_token"}
    example_account_info = {
        "user": {
            "email": existing_email,
        },
        "external_id": "1234",
        "external_method": "test_method",
    }

    # Mock remote app's handler
    current_oauthclient.signup_handlers[remote.name] = {
        "info": lambda resp: example_account_info,
    }

    _security.confirmable = True
    _security.login_without_confirmation = False
    user.confirmed_at = None

    expected_url_args = {"message": "Unauthorized.", "code": 401}
    resp = authorized_signup_handler(example_response, remote)
    check_response_redirect_url_args(resp, expected_url_args)


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_signup_handler(remote, app_rest, models_fixture):
    """Test signup handler."""
    datastore = app_rest.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)
    # Already authenticated
    login_user(user)
    assert current_user.is_authenticated
    resp1 = signup_handler(remote)
    expected_url_args = {"message": "Successfully signed up.", "code": 200}
    check_response_redirect_url_args(resp1, expected_url_args)
    logout_user()
    assert not current_user.is_authenticated

    # No OAuth token
    resp2 = signup_handler(remote)
    expected_url_args = {"message": "Token not found.", "code": 400}

    check_response_redirect_url_args(resp2, expected_url_args)

    # Not coming from authorized request
    token = RemoteToken.create(user.id, "testkey", "mytoken", "mysecret")
    token_setter(remote, token, "mysecret")
    with pytest.raises(BuildError):
        signup_handler(remote)


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_already_linked_exception(remote, app_rest):
    """Test error when service is already linked to another account."""

    @oauth_resp_remote_error_handler
    def mock_handler(resp, remote):
        raise AlreadyLinkedError(None, None)

    resp = mock_handler(None, remote)
    expected_url_args = {
        "message": "External service is already linked to another account.",
        "code": 400,
    }
    check_response_redirect_url_args(resp, expected_url_args)


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_unauthorized_disconnect(remote, app_rest):
    """Test disconnect handler when user is not authenticated."""

    def mock_unauthorized():
        return "Unauthorized"

    app_rest.login_manager.unauthorized = mock_unauthorized
    resp = disconnect_handler(remote)
    expected_url_args = {"message": "Unauthorized.", "code": 401}
    check_response_redirect_url_args(resp, expected_url_args)


@pytest.mark.parametrize("remote_name", REMOTE_APPS)
def test_dummy_handler(remote_name, base_app):
    """Test dummy handler."""

    signup_handler = base_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote_name][
        "signup_handler"
    ]

    # Force usage of dummy handlers
    base_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote_name]["signup_handler"] = {}

    # Initialize InvenioOAuth
    FlaskOAuth(base_app)
    InvenioOAuthClientREST(base_app)
    base_app.register_blueprint(rest_blueprint)

    # Try to sign-up client
    base_app.test_client().get(
        url_for(
            "invenio_oauthclient.rest_signup", remote_app=remote_name, next="/someurl/"
        )
    )

    base_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote_name][
        "signup_handler"
    ] = signup_handler


@pytest.mark.parametrize("remote_name", REMOTE_APPS)
def test_response_handler(remote_name, base_app):
    """Test response handler."""

    def mock_response_handler(remote, url, payload):
        return remote.name

    # Force usage of dummy handlers
    base_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote_name][
        "response_handler"
    ] = mock_response_handler

    # Initialize InvenioOAuth
    FlaskOAuth(base_app)
    InvenioOAuthClientREST(base_app)
    base_app.register_blueprint(rest_blueprint)

    # Try to sign-up client
    response = base_app.test_client().get(
        url_for("invenio_oauthclient.rest_signup", remote_app=remote_name)
    )
    assert remote_name in str(response.data)


@pytest.mark.parametrize("remote", REMOTE_APPS, indirect=["remote"])
def test_response_handler_with_postmessage(remote, base_app):
    """Test response handler with postmessage."""

    # Force usage of dummy handlers
    base_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name][
        "response_handler"
    ] = response_handler_postmessage

    # Initialize InvenioOAuth
    FlaskOAuth(base_app)
    InvenioOAuthClientREST(base_app)
    # The `rest_blueprint` is already registered indirectly by the
    # `remote` fixture

    datastore = base_app.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)
    # Already authenticated
    login_user(user)

    assert current_user.is_authenticated

    response = signup_handler(remote)
    expected_message = "Successfully signed up."
    expected_status = "200"

    assert expected_message in response
    assert expected_status in response
    assert "window.opener.postMessage" in response



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from flask import current_app, session
from flask_login import current_user
from invenio_accounts.models import Role
from invenio_accounts.proxies import current_datastore
from invenio_db import db

from ..errors import (
    OAuthClientAlreadyAuthorized,
    OAuthClientMustRedirectLogin,
    OAuthClientMustRedirectSignup,
    OAuthClientTokenNotFound,
    OAuthClientTokenNotSet,
    OAuthClientUnAuthorized,
    OAuthClientUserNotRegistered,
)
from ..models import RemoteAccount
from ..proxies import current_oauthclient
from ..signals import (
    account_info_received,
    account_setup_committed,
    account_setup_received,
)
from ..utils import (
    create_csrf_disabled_registrationform,
    fill_form,
    oauth_authenticate,
    oauth_get_user,
    oauth_register,
)
from .utils import (
    get_session_next_url,
    require_more_than_one_external_account,
    response_token_setter,
    token_getter,
    token_session_key,
    token_setter,
)


def _role_needs_update(role_obj, new_role_dict):
    """Checks if role needs to be updated."""
    if role_obj.name != new_role_dict.get(
        "name"
    ) or role_obj.description != new_role_dict.get("description"):
        return True
    return False


def create_or_update_groups(account_groups):
    """Creates the roles based on the groups provided."""
    roles_id = []
    for group in account_groups:
        existing_role = current_datastore.find_role_by_id(group["id"])
        if existing_role and existing_role.is_managed:
            current_app.logger.exception(
                f'Error while syncing roles: A managed role with id: ${group["id"]} already exists'
            )
            continue
        existing_role_by_name = current_datastore.find_role(group["name"])
        if existing_role_by_name and existing_role_by_name.is_managed:
            current_app.logger.exception(
                f'Error while syncing roles: A managed role with name: ${group["name"]} already exists'
            )
            continue
        if not existing_role:
            role = current_datastore.create_role(
                id=group["id"],
                name=group.get("name"),
                description=group.get("description"),
                is_managed=False,
            )
            roles_id.append(role.id)
        elif existing_role and _role_needs_update(existing_role, group):
            role_to_update = Role(
                id=group["id"],
                name=group.get("name"),
                description=group.get("description"),
                is_managed=False,
            )
            role = current_datastore.update_role(role_to_update)
            roles_id.append(role.id)
        else:
            roles_id.append(existing_role.id)

    current_datastore.commit()

    return roles_id


#
# Handlers
#
def base_authorized_signup_handler(resp, remote, *args, **kwargs):
    """Handle sign-in/up functionality.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: Redirect response.
    """
    # Remove any previously stored auto register session key
    session.pop(token_session_key(remote.name) + "_autoregister", None)
    # We set the remote in the session to be aware of which one is being used and, on log out redirect to
    # the correct URL set in the OAUTHCLIENT_REMOTE_APPS for each remote
    session["OAUTHCLIENT_SESSION_REMOTE_NAME"] = remote.name
    # Store token in session
    # ----------------------
    # Set token in session - token object only returned if
    # current_user.is_authenticated().
    token = response_token_setter(remote, resp)
    handlers = current_oauthclient.signup_handlers[remote.name]
    # Needed for tests
    if not current_user.is_authenticated:
        # Sign-in/up user
        # ---------------
        account_info = handlers["info"](resp)
        assert "external_id" in account_info
        account_info_received.send(
            remote, token=token, response=resp, account_info=account_info
        )

        user = oauth_get_user(
            remote.consumer_key,
            account_info=account_info,
            access_token=token_getter(remote)[0],
        )

        if user is None:
            # Auto sign-up if user not found
            form = create_csrf_disabled_registrationform(remote)
            form = fill_form(form, account_info["user"])
            remote_app = current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name]
            precedence_mask = remote_app.get("precedence_mask")
            signup_options = remote_app.get("signup_options")
            user = oauth_register(
                form,
                account_info["user"],
                precedence_mask=precedence_mask,
                signup_options=signup_options,
            )

            # if registration fails ...
            if user is None:
                # requires extra information
                session[token_session_key(remote.name) + "_autoregister"] = True
                session[token_session_key(remote.name) + "_account_info"] = account_info
                session[token_session_key(remote.name) + "_response"] = resp
                db.session.commit()
                raise OAuthClientMustRedirectSignup()

        group_handler = handlers.get("groups")
        if group_handler:
            account_groups = group_handler(resp)
            if account_groups:
                roles_id = create_or_update_groups(account_groups)
                # We set the unmanaged groups in the session because they are not linked to the user in the DB
                session["_unmanaged_groups"] = roles_id

        # Authenticate user after the unmanaged groups where set in the session
        if not oauth_authenticate(
            remote.consumer_key, user, require_existing_link=False
        ):
            raise OAuthClientUnAuthorized()

        # Link account
        # ------------
        # Need to store token in database instead of only the session when
        # called first time.
        token = response_token_setter(remote, resp)

    # Setup account
    # -------------
    if not token.remote_account.extra_data:
        account_setup = handlers["setup"](token, resp)
        account_setup_received.send(
            remote, token=token, response=resp, account_setup=account_setup
        )
        db.session.commit()
        account_setup_committed.send(remote, token=token)
    else:
        db.session.commit()

    # Redirect to next
    next_url = get_session_next_url(remote.name)
    if next_url:
        return next_url


@require_more_than_one_external_account
def base_disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    This default handler will just delete the remote account link. You may
    wish to extend this module to perform clean-up in the remote service
    before removing the link (e.g. removing install webhooks).

    :param remote: The remote application.
    :returns: Redirect response.
    """
    if not current_user.is_authenticated:
        raise OAuthClientUnAuthorized()

    with db.session.begin_nested():
        account = RemoteAccount.get(
            user_id=current_user.get_id(), client_id=remote.consumer_key
        )
        if account:
            account.delete()

    db.session.commit()


def base_signup_handler(remote, form, *args, **kwargs):
    """Handle extra signup information.

    :param remote: The remote application.
    :returns: Redirect response or the template rendered.
    """
    # User already authenticated so move on
    if current_user.is_authenticated:
        raise OAuthClientAlreadyAuthorized()

    # Retrieve token from session
    oauth_token = token_getter(remote)
    if not oauth_token:
        raise OAuthClientTokenNotFound()

    session_prefix = token_session_key(remote.name)

    # Test to see if this is coming from on authorized request
    if not session.get(session_prefix + "_autoregister", False):
        raise OAuthClientMustRedirectLogin()

    if form.validate_on_submit():
        account_info = session.get(session_prefix + "_account_info")
        response = session.get(session_prefix + "_response")

        remote_app = current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name]
        precedence_mask = remote_app.get("precedence_mask")
        signup_options = remote_app.get("signup_options")

        # Register user
        user = oauth_register(
            form,
            account_info["user"],
            precedence_mask=precedence_mask,
            signup_options=signup_options,
        )

        if user is None:
            raise OAuthClientUserNotRegistered()

        # Remove session key
        session.pop(session_prefix + "_autoregister", None)

        # Link account and set session data
        token = token_setter(remote, oauth_token[0], secret=oauth_token[1], user=user)
        handlers = current_oauthclient.signup_handlers[remote.name]

        if token is None:
            raise OAuthClientTokenNotSet()

        if not token.remote_account.extra_data:
            account_setup = handlers["setup"](token, response)
            account_setup_received.send(
                remote, token=token, response=response, account_setup=account_setup
            )
            # Registration has been finished
            db.session.commit()
            account_setup_committed.send(remote, token=token)
        else:
            # Registration has been finished
            db.session.commit()

        # Authenticate user
        if not oauth_authenticate(
            remote.consumer_key, user, require_existing_link=False
        ):
            raise OAuthClientUnAuthorized()

        # Remove account info from session
        session.pop(session_prefix + "_account_info", None)
        session.pop(session_prefix + "_response", None)

        # Redirect to next
        next_url = get_session_next_url(remote.name)
        if next_url:
            return next_url

    # Pre-fill form
    account_info = session.get(session_prefix + "_account_info")
    if not form.is_submitted():
        fill_form(form, account_info["user"])



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""OAuth client test utility functions."""

import json
from inspect import isfunction

import httpretty
import six
from mock import MagicMock
from six.moves.urllib_parse import parse_qs, urlencode, urlparse

from invenio_oauthclient._compat import _create_identifier
from invenio_oauthclient.views.client import serializer


def get_state(app="test"):
    """Get state."""
    return serializer.dumps(
        {
            "app": app,
            "sid": _create_identifier(),
            "next": None,
        }
    )


def mock_response(oauth, remote_app="test", data=None):
    """Mock the oauth response to use the remote."""
    oauth.remote_apps[remote_app].handle_oauth2_response = MagicMock(return_value=data)


def mock_remote_get(oauth, remote_app="test", data=None):
    """Mock the oauth remote get response."""
    oauth.remote_apps[remote_app].get = MagicMock(return_value=data)


def check_redirect_location(resp, loc):
    """Check response redirect location."""
    assert resp._status_code == 302
    if isinstance(loc, six.string_types):
        assert resp.headers["Location"] == loc
    elif isfunction(loc):
        assert loc(resp.headers["Location"])


def check_response_redirect_url(response, expected_url):
    """Check response redirect url."""
    assert response.status_code == 302
    state = serializer.loads(parse_qs(urlparse(response.location).query)["state"][0])
    assert expected_url == state["next"]


def check_response_redirect_url_args(response, expected_args):
    """Check response redirect url."""
    assert response.status_code == 302
    assert urlencode(expected_args) == urlparse(response.location).query


def mock_keycloak(app_config, token_dict, user_info_dict, realm_info):
    """Mock a running Keycloak instance."""
    app_config["OAUTHCLIENT_KEYCLOAK_USER_INFO_FROM_ENDPOINT"] = False
    keycloak_settings = app_config["OAUTHCLIENT_REMOTE_APPS"]["keycloak"]

    httpretty.register_uri(
        httpretty.POST,
        keycloak_settings["params"]["access_token_url"],
        body=json.dumps(token_dict),
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        app_config["OAUTHCLIENT_KEYCLOAK_USER_INFO_URL"],
        body=json.dumps(user_info_dict),
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        app_config["OAUTHCLIENT_KEYCLOAK_REALM_URL"],
        body=json.dumps(realm_info),
        content_type="application/json",
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Common OAuth configuration helper."""

from werkzeug import cached_property


class OAuthSettingsHelper:
    """Helper for creating REMOTE_APP configuration dictionaries for OAuth."""

    _remote_app = None
    _remote_rest_app = None

    def __init__(
        self,
        title,
        description,
        base_url,
        app_key,
        icon=None,
        access_token_url=None,
        authorize_url=None,
        access_token_method="POST",
        request_token_params=None,
        request_token_url=None,
        precedence_mask=None,
        signup_options=None,
        logout_url=None,
        **kwargs,
    ):
        """The constructor."""
        self.base_url = "{}/".format(base_url.rstrip("/"))  # add leading `/`
        icon = icon or ""
        request_token_params = request_token_params or {"scope": ""}
        access_token_url = access_token_url or f"{self.base_url}oauth2/token"
        authorize_url = authorize_url or f"{self.base_url}oauth2/authorize"
        precedence_mask = precedence_mask or {
            "email": True,
            "password": False,
            "profile": {
                "username": False,
                "full_name": False,
            },
        }

        signup_options = signup_options or {}
        signup_options.setdefault("auto_confirm", True)
        signup_options.setdefault("send_register_msg", False)

        self.base_app = dict(
            title=title,
            description=description,
            icon=icon,
            precedence_mask=precedence_mask,
            signup_options=signup_options,
            logout_url=logout_url,
            params=dict(
                base_url=self.base_url,
                request_token_params=request_token_params,
                request_token_url=request_token_url,
                access_token_url=access_token_url,
                access_token_method=access_token_method,
                authorize_url=authorize_url,
                app_key=app_key,
                **kwargs,
            ),
        )

    def get_handlers(self):
        """Return a dict with the auth handlers.

        It should return a dict in this format:
        dict(
            authorized_handler='path_to_method_authorized_signup_handler',
            disconnect_handler='path_to_method_authorized_disconnect_handler',
        signup_handler=dict(
            info='path_to_method_account_info',
            info_serializer='path_to_method_account_info_serializer',
            groups="path_to_method_account_groups_handler",
            groups_serializer="path_to_method_account_groups_serializer_handler",
            setup='path_to_method_account_setup',
            view='path_to_method_signup_form_handler',
        )
        """
        raise NotImplementedError

    @cached_property
    def remote_app(self):
        """Create a KEYCLOAK_REMOTE_APP using the given base URL and realm."""
        self._remote_app = dict(self.base_app)
        self._remote_app.update(self.get_handlers())
        return self._remote_app

    def get_rest_handlers(self):
        """Return a dict with the auth REST handlers.

        It should return a dict in this format:
        dict(
            authorized_handler='path_to_method_authorized_signup_handler',
            disconnect_handler='path_to_method_disconnect_rest_handler',
            signup_handler=dict(
                info='path_to_method_account_info',
                info_serializer='path_to_method_account_info_serializer',
                setup='path_to_method_account_setup',
                view='path_to_method_signup_form_handler',
                groups="path_to_method_account_groups_handler",
                groups_serializer="path_to_method_account_groups_serializer_handler",
            ),
            response_handler=(
                'path_to_method_response_handler'
            ),
            authorized_redirect_url='/',
            disconnect_redirect_url='/',
            signup_redirect_url='/',
            error_redirect_url='/'
        )
        """
        raise NotImplementedError

    @cached_property
    def remote_rest_app(self):
        """Crete a KEYCLOAK_REMOTE_REST_APP using the given configuration."""
        self._remote_rest_app = dict(self.base_app)
        self._remote_rest_app.update(self.get_rest_handlers())
        return self._remote_rest_app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Account settings blueprint for oauthclient."""

from operator import itemgetter

import six
from flask import Blueprint, current_app, render_template, request
from flask_breadcrumbs import register_breadcrumb
from flask_login import current_user, login_required
from flask_menu import register_menu
from invenio_i18n import lazy_gettext as _
from invenio_theme.proxies import current_theme_icons
from speaklater import make_lazy_string

from ..models import RemoteAccount
from ..proxies import current_oauthclient

blueprint = Blueprint(
    "invenio_oauthclient_settings",
    __name__,
    url_prefix="/account/settings/linkedaccounts",
    static_folder="../static",
    template_folder="../templates",
)


@blueprint.route("/", methods=["GET", "POST"])
@login_required
@register_menu(
    blueprint,
    "settings.oauthclient",
    _(
        "%(icon)s Linked accounts",
        icon=make_lazy_string(lambda: f'<i class="{current_theme_icons.link}"></i>'),
    ),
    order=3,
    active_when=lambda: request.endpoint.startswith("invenio_oauthclient_settings."),
    visible_when=lambda: bool(current_app.config.get("OAUTHCLIENT_REMOTE_APPS"))
    is not False,
)
@register_breadcrumb(
    blueprint, "breadcrumbs.settings.oauthclient", _("Linked accounts")
)
def index():
    """List linked accounts."""
    oauth = current_oauthclient.oauth

    services = []
    service_map = {}
    i = 0

    for appid, conf in six.iteritems(current_app.config["OAUTHCLIENT_REMOTE_APPS"]):
        if not conf.get("hide", False):
            services.append(
                dict(
                    appid=appid,
                    title=conf["title"],
                    icon=conf.get("icon", None),
                    description=conf.get("description", None),
                    account=None,
                )
            )
            service_map[oauth.remote_apps[appid].consumer_key] = i
            i += 1

    # Fetch already linked accounts
    accounts = RemoteAccount.query.filter_by(user_id=current_user.get_id()).all()

    for a in accounts:
        if a.client_id in service_map:
            services[service_map[a.client_id]]["account"] = a

    # Sort according to title
    services.sort(key=itemgetter("title"))

    # Check if local login is possible
    local_login_enabled = current_app.config.get("ACCOUNTS_LOCAL_LOGIN_ENABLED", True)
    password_set = current_user.password is not None
    local_login_possible = local_login_enabled and password_set

    return render_template(
        "invenio_oauthclient/settings/index.html",
        services=services,
        only_external_login=not local_login_possible,
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pre-configured remote application for enabling sign in/up with GitHub.

1. Ensure you have ``github3.py`` package installed:

   .. code-block:: console

      cdvirtualenv src/invenio-oauthclient
      pip install -e .[github]

2. Edit your configuration and add:

   .. code-block:: python

        from invenio_oauthclient.contrib import github

        OAUTHCLIENT_REMOTE_APPS = dict(
            github=github.REMOTE_APP,
        )

        GITHUB_APP_CREDENTIALS = dict(
            consumer_key='changeme',
            consumer_secret='changeme',
        )

3. Go to GitHub and register a new application:
   https://github.com/settings/applications/new. When registering the
   application ensure that the *Authorization callback URL* points to:
   ``CFG_SITE_SECURE_URL/oauth/authorized/github/`` (e.g.
   ``http://localhost:4000/oauth/authorized/github/`` for development).


4. Grab the *Client ID* and *Client Secret* after registering the application
   and add them to your instance configuration (``invenio.cfg``):

   .. code-block:: python

        GITHUB_APP_CREDENTIALS = dict(
            consumer_key='<CLIENT ID>',
            consumer_secret='<CLIENT SECRET>',
        )

5. Now go to ``CFG_SITE_SECURE_URL/oauth/login/github/`` (e.g.
   http://localhost:4000/oauth/login/github/)

6. Also, you should see GitHub listed under Linked accounts:
   http://localhost:4000/account/settings/linkedaccounts/

By default the GitHub module will try first look if a link already exists
between a GitHub account and a user. If no link is found, the module tries to
retrieve the user email address from GitHub to match it with a local user. If
this fails, the user is asked to provide an email address to sign-up.

In templates you can add a sign in/up link:

.. code-block:: jinja

    <a href='{{url_for('invenio_oauthclient.login', remote_app='github')}}'>
      Sign in with GitHub
    </a>

For more details you can play with a :doc:`working example <examplesapp>`.
"""

import github3
from flask import current_app, redirect, url_for
from flask_login import current_user
from invenio_db import db

from invenio_oauthclient import current_oauthclient
from invenio_oauthclient.contrib.settings import OAuthSettingsHelper
from invenio_oauthclient.errors import OAuthResponseError
from invenio_oauthclient.handlers import authorized_signup_handler, oauth_error_handler
from invenio_oauthclient.handlers.rest import (
    authorized_signup_handler as authorized_signup_rest_handler,
)
from invenio_oauthclient.handlers.rest import (
    oauth_resp_remote_error_handler,
    response_handler,
)
from invenio_oauthclient.handlers.utils import require_more_than_one_external_account
from invenio_oauthclient.models import RemoteAccount
from invenio_oauthclient.utils import oauth_link_external_id, oauth_unlink_external_id


class GitHubOAuthSettingsHelper(OAuthSettingsHelper):
    """Default configuration for GitHub OAuth provider."""

    def __init__(
        self,
        title=None,
        description=None,
        base_url=None,
        app_key=None,
        icon=None,
        precedence_mask=None,
        signup_options=None,
    ):
        """Constructor."""
        super().__init__(
            title or "GitHub",
            description or "Software collaboration platform.",
            base_url or "https://api.github.com/",
            app_key or "GITHUB_APP_CREDENTIALS",
            icon=icon or "fa fa-github",
            request_token_params={"scope": "user,user:email"},
            access_token_url="https://github.com/login/oauth/access_token",
            authorize_url="https://github.com/login/oauth/authorize",
            precedence_mask=precedence_mask,
            signup_options=signup_options,
        )

        self._handlers = dict(
            authorized_handler="invenio_oauthclient.handlers:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.github:disconnect_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.github:account_info",
                info_serializer="invenio_oauthclient.contrib.github:account_info_serializer",
                setup="invenio_oauthclient.contrib.github:account_setup",
                view="invenio_oauthclient.handlers:signup_handler",
            ),
        )

        self._rest_handlers = dict(
            authorized_handler="invenio_oauthclient.handlers.rest"
            ":authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.github"
            ":disconnect_rest_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.github:account_info",
                info_serializer="invenio_oauthclient.contrib.github:account_info_serializer",
                setup="invenio_oauthclient.contrib.github:account_setup",
                view="invenio_oauthclient.handlers.rest:signup_handler",
            ),
            response_handler="invenio_oauthclient.handlers.rest"
            ":default_remote_response_handler",
            authorized_redirect_url="/",
            disconnect_redirect_url="/",
            signup_redirect_url="/",
            error_redirect_url="/",
        )

    def get_handlers(self):
        """Return GitHub auth handlers."""
        return self._handlers

    def get_rest_handlers(self):
        """Return GitHub auth REST handlers."""
        return self._rest_handlers


_github_app = GitHubOAuthSettingsHelper()

BASE_APP = _github_app.base_app
REMOTE_APP = _github_app.remote_app
"""GitHub remote application configuration."""
REMOTE_REST_APP = _github_app.remote_rest_app
"""GitHub remote rest application configuration."""


def _extract_email(emails):
    """Get user email from GitHub."""
    return next((x.email for x in emails if x.verified and x.primary), None)


def account_info_serializer(remote, resp, user_info, emails_info, **kwargs):
    """Serialize the account info response object.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :param user_info: The response of the `me` endpoint.
    :param emails_info: The response of the `emails` endpoint.
    :returns: A dictionary with serialized user information.
    """
    return dict(
        user=dict(
            email=_extract_email(emails_info),
            profile=dict(
                username=user_info.login,
                full_name=user_info.name,
            ),
        ),
        external_id=str(user_info.id),
        external_method=remote.name,
    )


def account_info(remote, resp):
    """Retrieve remote account information used to find local user.

    It returns a dictionary with the following structure:

    .. code-block:: python

        {
            'user': {
                'email': '...',
                'profile': {
                    'username': '...',
                    'full_name': '...',
                }
            },
            'external_id': 'github-unique-identifier',
            'external_method': 'github',
        }

    Information inside the user dictionary are available for other modules.
    For example, they are used from the module invenio-userprofiles to fill
    the user profile.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: A dictionary with the user information.
    """
    github_obj = github3.login(token=resp["access_token"])
    user_info = github_obj.me()
    emails_info = github_obj.emails()

    handlers = current_oauthclient.signup_handlers[remote.name]
    # `remote` param automatically injected via `make_handler` helper
    return handlers["info_serializer"](resp, user_info, emails_info=emails_info)


def account_setup(remote, token, resp):
    """Perform additional setup after user have been logged in.

    :param remote: The remote application.
    :param token: The token value.
    :param resp: The response.
    """
    gh = github3.login(token=resp["access_token"])
    with db.session.begin_nested():
        me = gh.me()

        token.remote_account.extra_data = {"login": me.login, "id": me.id}

        # Create user <-> external id link.
        oauth_link_external_id(
            token.remote_account.user, dict(id=str(me.id), method=remote.name)
        )


@oauth_error_handler
def authorized(resp, remote):
    """Authorized callback handler for GitHub.

    :param resp: The response.
    :param remote: The remote application.
    """
    if resp and "error" in resp:
        if resp["error"] == "bad_verification_code":
            # See https://developer.github.com/v3/oauth/#bad-verification-code
            # which recommends starting auth flow again.
            return redirect(url_for("invenio_oauthclient.login", remote_app="github"))
        elif resp["error"] in ["incorrect_client_credentials", "redirect_uri_mismatch"]:
            raise OAuthResponseError(
                "Application mis-configuration in GitHub", remote, resp
            )

    return authorized_signup_handler(resp, remote)


@oauth_resp_remote_error_handler
def authorized_rest(resp, remote):
    """Authorized callback handler for GitHub.

    :param resp: The response.
    :param remote: The remote application.
    """
    if resp and "error" in resp:
        if resp["error"] == "bad_verification_code":
            # See https://developer.github.com/v3/oauth/#bad-verification-code
            # which recommends starting auth flow again.
            return redirect(
                url_for("invenio_oauthclient.rest_login", remote_app="github")
            )
        elif resp["error"] in ["incorrect_client_credentials", "redirect_uri_mismatch"]:
            raise OAuthResponseError(
                "Application mis-configuration in GitHub", remote, resp
            )

    return authorized_signup_rest_handler(resp, remote)


@require_more_than_one_external_account
def _disconnect(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    :returns: The HTML response.
    """
    if not current_user.is_authenticated:
        return current_app.login_manager.unauthorized()

    remote_account = RemoteAccount.get(
        user_id=current_user.get_id(), client_id=remote.consumer_key
    )
    external_ids = [
        i.id for i in current_user.external_identifiers if i.method == remote.name
    ]

    if external_ids:
        oauth_unlink_external_id(dict(id=external_ids[0], method=remote.name))
    if remote_account:
        with db.session.begin_nested():
            remote_account.delete()


def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    :returns: The HTML response.
    """
    _disconnect(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


def disconnect_rest_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    :returns: The HTML response.
    """
    _disconnect(remote, *args, **kwargs)
    redirect_url = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name][
        "disconnect_redirect_url"
    ]
    return response_handler(remote, redirect_url)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for models."""

from invenio_db import db

from invenio_oauthclient.models import RemoteAccount, RemoteToken


def test_get_create_remote_account(app, models_fixture):
    """Test create remote account."""
    created_acc = RemoteAccount.create(1, "dev", dict(somekey="somevalue"))
    assert created_acc

    retrieved_acc = RemoteAccount.get(1, "dev")
    assert created_acc.id == retrieved_acc.id
    assert retrieved_acc.extra_data == dict(somekey="somevalue")

    db.session.delete(retrieved_acc)
    assert RemoteAccount.get(1, "dev") is None


def test_get_create_remote_token(app, models_fixture):
    """Test create remote token."""
    existing_email = "existing@inveniosoftware.org"
    datastore = app.extensions["invenio-accounts"].datastore
    user = datastore.find_user(email=existing_email)

    t = RemoteToken.create(user.id, "dev", "mytoken", "mysecret")
    assert t
    assert t.token() == ("mytoken", "mysecret")

    acc = RemoteAccount.get(user.id, "dev")
    assert acc
    assert t.remote_account.id == acc.id
    assert t.token_type == ""

    t2 = RemoteToken.create(user.id, "dev", "mytoken2", "mysecret2", token_type="t2")
    assert t2.remote_account.id == acc.id
    assert t2.token_type == "t2"

    t3 = RemoteToken.get(user.id, "dev")
    t4 = RemoteToken.get(user.id, "dev", token_type="t2")
    assert t4.token() != t3.token()

    assert RemoteToken.query.count() == 2
    acc.delete()
    assert RemoteToken.query.count() == 0


def test_get_regression(app, models_fixture):
    """Test regression."""
    datastore = app.extensions["invenio-accounts"].datastore

    email2 = "test2@inveniosoftware.org"
    email3 = "test3@inveniosoftware.org"

    user2 = datastore.find_user(email=email2)
    user3 = datastore.find_user(email=email3)

    t3 = RemoteToken.create(user2.id, "dev", "mytoken", "mysecret")
    t4 = RemoteToken.create(user3.id, "dev", "mytoken", "mysecret")

    assert (
        RemoteToken.get(user2.id, "dev").remote_account.user_id
        == t3.remote_account.user_id
    )
    assert (
        RemoteToken.get(user3.id, "dev").remote_account.user_id
        == t4.remote_account.user_id
    )


def test_repr(app, models_fixture):
    """Test representation of RemoteAccount and RemoteToken."""
    datastore = app.extensions["invenio-accounts"].datastore
    user = datastore.find_user(email="existing@inveniosoftware.org")

    assert "Remote Token <token_type=type access_token=****oken>" == repr(
        RemoteToken.create(user.id, "dev", "mytoken", "mysecret", token_type="type")
    )

    assert "Remote Account <id=1, user_id=1>" == repr(RemoteAccount.get(user.id, "dev"))
